//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      Base/Types/OwningVector.h
//! @brief     Defines and implements templated class OwningVector.
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2018
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#ifdef SWIG
#error no need to expose this header to Swig
#endif

#ifndef USER_API
#ifndef BORNAGAIN_BASE_TYPES_OWNINGVECTOR_H
#define BORNAGAIN_BASE_TYPES_OWNINGVECTOR_H

#include <cstddef>
#include <utility>
#include <vector>

//! A vector of unique pointers to objects that are cloneable.
//!
//! Equips vector<unique_ptr<T>> with copy constructor.
//! For use with polymorphic objects, or in pimpl idiom.

//! The objects pointed to must posses a clone() function.

template <class T>
class OwningVector : private std::vector<T*> {
    using super = std::vector<T*>;

public:
    OwningVector() = default;
    ~OwningVector() { clear(); }
    OwningVector(const OwningVector& other)
        : super()
    {
        super::reserve(other.size());
        for (T* e : other)
            super::emplace_back(e->clone());
    }
    OwningVector& operator=(const OwningVector& other)
    {
        if (this == &other)
            return *this;
        OwningVector ret(other);
        std::swap(*this, ret);
        return *this;
    }

    void emplace_back(T* e) { super::emplace_back(e); }
    void clear()
    {
        for (T* e : *this)
            delete e;
        super::clear();
    }

    size_t size() const { return super::size(); }
    bool empty() const { return super::empty(); }
    T* const& operator[](int i) const { return super::operator[](i); }
    T* const& at(int i) const { return super::at(i); }
    const T* back() const { return super::back(); }

    using Iterator = typename super::iterator; // "typename" can be dropped under C++20
    using ConstIterator = typename super::const_iterator;

    ConstIterator begin() const { return super::begin(); }
    ConstIterator end() const { return super::end(); }
    Iterator begin() { return super::begin(); }
    Iterator end() { return super::end(); }
};

#endif // BORNAGAIN_BASE_TYPES_OWNINGVECTOR_H
#endif // USER_API
